/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2015-2025 Advanced Micro Devices, Inc. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#include <migraphx/migraphx.hpp>
#include "test.hpp"

struct array2 : migraphx::array_base<array2>
{
    std::vector<int> v;
    array2() = default;
    array2(std::initializer_list<int> x) : v(x) {}
    std::size_t size() const { return v.size(); }
    int operator[](std::size_t i) const { return v[i]; }
};

TEST_CASE(iterators)
{
    array2 a = {1, 2, 3};
    EXPECT(std::equal(a.begin(), a.end(), a.v.begin()));
}

TEST_CASE(front_back)
{
    array2 a = {1, 2, 3};
    EXPECT(a.front() == 1);
    EXPECT(a.back() == 3);
}

TEST_CASE(empty)
{
    array2 a = {1, 2, 3};
    EXPECT(not a.empty());
}

int main(int argc, const char* argv[]) { test::run(argc, argv); }
